@@ -12,6 +12,8 @@
 #include <algorithm>
 #include <map>
 #include <memory>
+#include <ostream>
+#include <sstream>
 #include <string>
 #include <vector>
 #include <utility>
@@ -27,16 +29,20 @@
 
 #define TREELITE_MAX_PRED_TRANSFORM_LENGTH 256
 
-/* Foward declarations */
-namespace dmlc {
-
-class Stream;
-float stof(const std::string& value, std::size_t* pos);
-
-}  // namespace dmlc
+/* Indicator that certain functions should be visible from a library (Windows only) */
+#if defined(_MSC_VER) || defined(_WIN32)
+#define TREELITE_DLL_EXPORT __declspec(dllexport)
+#else
+#define TREELITE_DLL_EXPORT
+#endif
 
 namespace treelite {
 
+class GTILBridge;
+
+template <typename ThresholdType, typename LeafOutputType>
+class ModelImpl;
+
 // Represent a frame in the Python buffer protocol (PEP 3118). We use a simplified representation
 // to hold only 1-D arrays with stride 1.
 struct PyBufferFrame {
@@ -67,6 +73,7 @@ class ContiguousArray {
   inline T& Back();
   inline const T& Back() const;
   inline std::size_t Size() const;
+  inline bool Empty() const;
   inline void Reserve(std::size_t newsize);
   inline void Resize(std::size_t newsize);
   inline void Resize(std::size_t newsize, T t);
@@ -160,8 +167,18 @@ enum class TaskType : uint8_t {
   kMultiClfCategLeaf = 3
 };
 
+inline std::string TaskTypeToString(TaskType type) {
+  switch (type) {
+    case TaskType::kBinaryClfRegr: return "BinaryClfRegr";
+    case TaskType::kMultiClfGrovePerClass: return "MultiClfGrovePerClass";
+    case TaskType::kMultiClfProbDistLeaf: return "MultiClfProbDistLeaf";
+    case TaskType::kMultiClfCategLeaf: return "MultiClfCategLeaf";
+    default: return "";
+  }
+}
+
 /*! \brief Group of parameters that are dependent on the choice of the task type. */
-struct TaskParameter {
+struct TaskParam {
   enum class OutputType : uint8_t { kFloat = 0, kInt = 1 };
   /*! \brief The type of output from each leaf node. */
   OutputType output_type;
@@ -190,7 +207,15 @@ struct TaskParameter {
   unsigned int leaf_vector_size;
 };
 
-static_assert(std::is_pod<TaskParameter>::value, "TaskParameter must be POD type");
+inline std::string OutputTypeToString(TaskParam::OutputType type) {
+  switch (type) {
+    case TaskParam::OutputType::kFloat: return "float";
+    case TaskParam::OutputType::kInt: return "int";
+    default: return "";
+  }
+}
+
+static_assert(std::is_pod<TaskParam>::value, "TaskParameter must be POD type");
 
 /*! \brief in-memory representation of a decision tree */
 template <typename ThresholdType, typename LeafOutputType>
@@ -207,19 +232,19 @@ class Tree {
       ThresholdType threshold;   // for non-leaf nodes
     };
     /*! \brief pointer to left and right children */
-    int32_t cleft_, cright_;
+    std::int32_t cleft_, cright_;
     /*!
      * \brief feature index used for the split
      * highest bit indicates default direction for missing values
      */
-    uint32_t sindex_;
+    std::uint32_t sindex_;
     /*! \brief storage for leaf value or decision threshold */
     Info info_;
     /*!
      * \brief number of data points whose traversal paths include this node.
      *        LightGBM models natively store this statistics.
      */
-    uint64_t data_count_;
+    std::uint64_t data_count_;
     /*!
      * \brief sum of hessian values for all data points whose traversal paths
      *        include this node. This value is generally correlated positively
@@ -248,6 +273,62 @@ class Tree {
     /* \brief whether the list given by MatchingCategories(nid) is associated with the right child
      *        node or the left child node. True if the right child, False otherwise */
     bool categories_list_right_child_;
+
+    /** Getters **/
+    inline int LeftChild() const {
+      return cleft_;
+    }
+    inline int RightChild() const {
+      return cright_;
+    }
+    inline bool DefaultLeft() const {
+      // Extract the most significant bit (MSB) of sindex_, which encodes the default_left field
+      return (sindex_ >> 31U) != 0;
+    }
+    inline int DefaultChild() const {
+      // Extract the most significant bit (MSB) of sindex_, which encodes the default_left field
+      return ((sindex_ >> 31U) != 0) ? cleft_ : cright_;
+    }
+    inline std::uint32_t SplitIndex() const {
+      // Extract all bits except the most significant bit (MSB) from sindex_.
+      return (sindex_ & ((1U << 31U) - 1U));
+    }
+    inline bool IsLeaf() const {
+      return cleft_ == -1;
+    }
+    inline LeafOutputType LeafValue() const {
+      return info_.leaf_value;
+    }
+    inline ThresholdType Threshold() const {
+      return info_.threshold;
+    }
+    inline Operator ComparisonOp() const {
+      return cmp_;
+    }
+    inline SplitFeatureType SplitType() const {
+      return split_type_;
+    }
+    inline bool HasDataCount() const {
+      return data_count_present_;
+    }
+    inline std::uint64_t DataCount() const {
+      return data_count_;
+    }
+    inline bool HasSumHess() const {
+      return sum_hess_present_;
+    }
+    inline double SumHess() const {
+      return sum_hess_;
+    }
+    inline bool HasGain() const {
+      return gain_present_;
+    }
+    inline double Gain() const {
+      return gain_;
+    }
+    inline bool CategoriesListRightChild() const {
+      return categories_list_right_child_;
+    }
   };
 
   static_assert(std::is_pod<Node>::value, "Node must be a POD type");
@@ -265,7 +346,7 @@ class Tree {
                 || (std::is_same<ThresholdType, double>::value && sizeof(Node) == 56),
                 "Node size incorrect");
 
-  Tree() = default;
+  explicit Tree(bool use_opt_field = true);
   ~Tree() = default;
   Tree(const Tree&) = delete;
   Tree& operator=(const Tree&) = delete;
@@ -277,17 +358,36 @@ class Tree {
   inline const char* GetFormatStringForNode();
   inline void GetPyBuffer(std::vector<PyBufferFrame>* dest);
   inline void SerializeToFile(FILE* dest_fp);
-  inline void InitFromPyBuffer(std::vector<PyBufferFrame>::iterator begin,
-                               std::vector<PyBufferFrame>::iterator end);
+  // Load a Tree object from a sequence of PyBuffer frames
+  // Returns the updated position of the cursor in the sequence
+  inline std::vector<PyBufferFrame>::iterator
+    InitFromPyBuffer(std::vector<PyBufferFrame>::iterator it);
   inline void DeserializeFromFile(FILE* src_fp);
 
  private:
   // vector of nodes
   ContiguousArray<Node> nodes_;
   ContiguousArray<LeafOutputType> leaf_vector_;
-  ContiguousArray<std::size_t> leaf_vector_offset_;
-  ContiguousArray<uint32_t> matching_categories_;
+  // Map nid to the start and end index in leaf_vector_
+  // We could use std::pair, but it is not POD, so easier to use two vectors
+  // here
+  ContiguousArray<std::size_t> leaf_vector_begin_;
+  ContiguousArray<std::size_t> leaf_vector_end_;
+  ContiguousArray<std::uint32_t> matching_categories_;
   ContiguousArray<std::size_t> matching_categories_offset_;
+  bool has_categorical_split_{false};
+
+  /* Note: the following member fields shall be re-computed at serialization time */
+  // Whether to use optional fields
+  bool use_opt_field_{false};
+  // Number of optional fields in the extension slots
+  int32_t num_opt_field_per_tree_{0};
+  int32_t num_opt_field_per_node_{0};
+
+  template <typename WriterType, typename X, typename Y>
+  friend void DumpModelAsJSON(WriterType& writer, const ModelImpl<X, Y>& model);
+  template <typename WriterType, typename X, typename Y>
+  friend void DumpTreeAsJSON(WriterType& writer, const Tree<X, Y>& tree);
 
   // allocate a new node
   inline int AllocNode();
@@ -297,13 +397,16 @@ class Tree {
   inline void
   SerializeTemplate(ScalarHandler scalar_handler, PrimitiveArrayHandler primitive_array_handler,
       CompositeArrayHandler composite_array_handler);
-  template <typename ScalarHandler, typename ArrayHandler>
+  template <typename ScalarHandler, typename ArrayHandler, typename SkipOptFieldHandlerFunc>
   inline void
-  DeserializeTemplate(ScalarHandler scalar_handler, ArrayHandler array_handler);
+  DeserializeTemplate(ScalarHandler scalar_handler, ArrayHandler array_handler,
+      SkipOptFieldHandlerFunc skip_opt_field_handler);
+
+  friend class GTILBridge;  // bridge to enable optimized access to nodes from GTIL
 
  public:
   /*! \brief number of nodes */
-  int num_nodes;
+  int num_nodes{0};
   /*! \brief initialize the model with a single root node */
   inline void Init();
   /*!
@@ -312,69 +415,63 @@ class Tree {
    */
   inline void AddChilds(int nid);
 
-  /*!
-   * \brief get list of all categorical features that have appeared anywhere in tree
-   * \return list of all categorical features used
-   */
-  inline std::vector<unsigned> GetCategoricalFeatures() const;
-
   /** Getters **/
   /*!
    * \brief index of the node's left child
    * \param nid ID of node being queried
    */
   inline int LeftChild(int nid) const {
-    return nodes_.at(nid).cleft_;
+    return nodes_[nid].LeftChild();
   }
   /*!
    * \brief index of the node's right child
    * \param nid ID of node being queried
    */
   inline int RightChild(int nid) const {
-    return nodes_.at(nid).cright_;
+    return nodes_[nid].RightChild();
   }
   /*!
    * \brief index of the node's "default" child, used when feature is missing
    * \param nid ID of node being queried
    */
   inline int DefaultChild(int nid) const {
-    return DefaultLeft(nid) ? LeftChild(nid) : RightChild(nid);
+    return nodes_[nid].DefaultChild();
   }
   /*!
    * \brief feature index of the node's split condition
    * \param nid ID of node being queried
    */
-  inline uint32_t SplitIndex(int nid) const {
-    return (nodes_.at(nid).sindex_ & ((1U << 31U) - 1U));
+  inline std::uint32_t SplitIndex(int nid) const {
+    return nodes_[nid].SplitIndex();
   }
   /*!
    * \brief whether to use the left child node, when the feature in the split condition is missing
    * \param nid ID of node being queried
    */
   inline bool DefaultLeft(int nid) const {
-    return (nodes_.at(nid).sindex_ >> 31U) != 0;
+    return nodes_[nid].DefaultLeft();
   }
   /*!
    * \brief whether the node is leaf node
    * \param nid ID of node being queried
    */
   inline bool IsLeaf(int nid) const {
-    return nodes_.at(nid).cleft_ == -1;
+    return nodes_[nid].IsLeaf();
   }
   /*!
    * \brief get leaf value of the leaf node
    * \param nid ID of node being queried
    */
   inline LeafOutputType LeafValue(int nid) const {
-    return (nodes_.at(nid).info_).leaf_value;
+    return nodes_[nid].LeafValue();
   }
   /*!
    * \brief get leaf vector of the leaf node; useful for multi-class random forest classifier
    * \param nid ID of node being queried
    */
   inline std::vector<LeafOutputType> LeafVector(int nid) const {
-    const std::size_t offset_begin = leaf_vector_offset_.at(nid);
-    const std::size_t offset_end = leaf_vector_offset_.at(nid + 1);
+    const std::size_t offset_begin = leaf_vector_begin_[nid];
+    const std::size_t offset_end = leaf_vector_end_[nid];
     if (offset_begin >= leaf_vector_.Size() || offset_end > leaf_vector_.Size()) {
       // Return empty vector, to indicate the lack of leaf vector
       return std::vector<LeafOutputType>();
@@ -389,21 +486,21 @@ class Tree {
    * \param nid ID of node being queried
    */
   inline bool HasLeafVector(int nid) const {
-    return leaf_vector_offset_.at(nid) != leaf_vector_offset_.at(nid + 1);
+    return leaf_vector_begin_[nid] != leaf_vector_end_[nid];
   }
   /*!
    * \brief get threshold of the node
    * \param nid ID of node being queried
    */
   inline ThresholdType Threshold(int nid) const {
-    return (nodes_.at(nid).info_).threshold;
+    return nodes_[nid].Threshold();
   }
   /*!
    * \brief get comparison operator
    * \param nid ID of node being queried
    */
   inline Operator ComparisonOp(int nid) const {
-    return nodes_.at(nid).cmp_;
+    return nodes_[nid].ComparisonOp();
   }
   /*!
    * \brief Get list of all categories belonging to the left/right child node. See the
@@ -413,47 +510,39 @@ class Tree {
    *        assumed to be in ascending order.
    * \param nid ID of node being queried
    */
-  inline std::vector<uint32_t> MatchingCategories(int nid) const {
-    const std::size_t offset_begin = matching_categories_offset_.at(nid);
-    const std::size_t offset_end = matching_categories_offset_.at(nid + 1);
+  inline std::vector<std::uint32_t> MatchingCategories(int nid) const {
+    const std::size_t offset_begin = matching_categories_offset_[nid];
+    const std::size_t offset_end = matching_categories_offset_[nid + 1];
     if (offset_begin >= matching_categories_.Size() || offset_end > matching_categories_.Size()) {
       // Return empty vector, to indicate the lack of any matching categories
       // The node might be a numerical split
-      return std::vector<uint32_t>();
+      return std::vector<std::uint32_t>();
     }
-    return std::vector<uint32_t>(&matching_categories_[offset_begin],
-                                 &matching_categories_[offset_end]);
+    return std::vector<std::uint32_t>(&matching_categories_[offset_begin],
+                                      &matching_categories_[offset_end]);
       // Use unsafe access here, since we may need to take the address of one past the last
       // element, to follow with the range semantic of std::vector<>.
   }
-  /*!
-   * \brief tests whether the node has a non-empty list for matching categories. See
-   *        MatchingCategories() for the definition of matching categories.
-   * \param nid ID of node being queried
-   */
-  inline bool HasMatchingCategories(int nid) const {
-    return matching_categories_offset_.at(nid) != matching_categories_offset_.at(nid + 1);
-  }
   /*!
    * \brief get feature split type
    * \param nid ID of node being queried
    */
   inline SplitFeatureType SplitType(int nid) const {
-    return nodes_.at(nid).split_type_;
+    return nodes_[nid].SplitType();
   }
   /*!
    * \brief test whether this node has data count
    * \param nid ID of node being queried
    */
   inline bool HasDataCount(int nid) const {
-    return nodes_.at(nid).data_count_present_;
+    return nodes_[nid].HasDataCount();
   }
   /*!
    * \brief get data count
    * \param nid ID of node being queried
    */
-  inline uint64_t DataCount(int nid) const {
-    return nodes_.at(nid).data_count_;
+  inline std::uint64_t DataCount(int nid) const {
+    return nodes_[nid].DataCount();
   }
 
   /*!
@@ -461,28 +550,28 @@ class Tree {
    * \param nid ID of node being queried
    */
   inline bool HasSumHess(int nid) const {
-    return nodes_.at(nid).sum_hess_present_;
+    return nodes_[nid].HasSumHess();
   }
   /*!
    * \brief get hessian sum
    * \param nid ID of node being queried
    */
   inline double SumHess(int nid) const {
-    return nodes_.at(nid).sum_hess_;
+    return nodes_[nid].SumHess();
   }
   /*!
    * \brief test whether this node has gain value
    * \param nid ID of node being queried
    */
   inline bool HasGain(int nid) const {
-    return nodes_.at(nid).gain_present_;
+    return nodes_[nid].HasGain();
   }
   /*!
    * \brief get gain value
    * \param nid ID of node being queried
    */
   inline double Gain(int nid) const {
-    return nodes_.at(nid).gain_;
+    return nodes_[nid].Gain();
   }
   /*!
    * \brief test whether the list given by MatchingCategories(nid) is associated with the right
@@ -490,7 +579,14 @@ class Tree {
    * \param nid ID of node being queried
    */
   inline bool CategoriesListRightChild(int nid) const {
-    return nodes_.at(nid).categories_list_right_child_;
+    return nodes_[nid].CategoriesListRightChild();
+  }
+
+  /*!
+   * \brief Query whether this tree contains any categorical splits
+   */
+  inline bool HasCategoricalSplit() const {
+    return has_categorical_split_;
   }
 
   /** Setters **/
@@ -562,8 +658,6 @@ class Tree {
     node.gain_ = gain;
     node.gain_present_ = true;
   }
-
-  void ReferenceSerialize(dmlc::Stream* fo) const;
 };
 
 struct ModelParam {
@@ -597,6 +691,14 @@ struct ModelParam {
    * It must be strictly positive; if unspecified, it is set to 1.0.
    */
   float sigmoid_alpha;
+  /*!
+   * \brief scaling parameter for exponential standard ratio transformation
+   * `expstdratio(x) = exp2(-x / c)`
+   *
+   * This parameter is used only when `pred_transform` is set to `'exponential_standard_ratio'`.
+   * If unspecified, it is set to 1.0.
+   */
+  float ratio_c;
   /*!
    * \brief global bias of the model
    *
@@ -606,7 +708,7 @@ struct ModelParam {
   float global_bias;
   /*! \} */
 
-  ModelParam() : sigmoid_alpha(1.0f), global_bias(0.0f) {
+  ModelParam() : sigmoid_alpha(1.0f), ratio_c(1.0f), global_bias(0.0f) {
     std::memset(pred_transform, 0, TREELITE_MAX_PRED_TRANSFORM_LENGTH * sizeof(char));
     std::strncpy(pred_transform, "identity", sizeof(pred_transform));
   }
@@ -656,11 +758,29 @@ class Model {
 
   virtual std::size_t GetNumTree() const = 0;
   virtual void SetTreeLimit(std::size_t limit) = 0;
-  virtual void ReferenceSerialize(dmlc::Stream* fo) const = 0;
+  virtual void DumpAsJSON(std::ostream& fo, bool pretty_print) const = 0;
+
+  inline std::string DumpAsJSON(bool pretty_print) const {
+    std::ostringstream oss;
+    DumpAsJSON(oss, pretty_print);
+    return oss.str();
+  }
+
+  /* Compatibility Matrix:
+     +------------------+----------+----------+----------------+-----------+
+     |                  | To: =2.4 | To: =3.0 | To: >=3.1,<4.0 | To: >=4.0 |
+     +------------------+----------+----------+----------------+-----------+
+     | From: <2.4       | No       | No       | No             | No        |
+     | From: =2.4       | Yes      | Yes      | Yes            | No        |
+     | From: =3.0       | No       | Yes      | Yes            | Yes       |
+     | From: >=3.1,<4.0 | No       | Yes      | Yes            | Yes       |
+     | From: >=4.0      | No       | No       | No             | Yes       |
+     +------------------+----------+----------+----------------+-----------+ */
 
   /* In-memory serialization, zero-copy */
-  std::vector<PyBufferFrame> GetPyBuffer();
-  static std::unique_ptr<Model> CreateFromPyBuffer(std::vector<PyBufferFrame> frames);
+  TREELITE_DLL_EXPORT std::vector<PyBufferFrame> GetPyBuffer();
+  TREELITE_DLL_EXPORT static std::unique_ptr<Model>
+    CreateFromPyBuffer(std::vector<PyBufferFrame> frames);
 
   /* Serialization to a file stream */
   void SerializeToFile(FILE* dest_fp);
@@ -670,31 +790,44 @@ class Model {
    * \brief number of features used for the model.
    * It is assumed that all feature indices are between 0 and [num_feature]-1.
    */
-  int num_feature;
+  int32_t num_feature{0};
   /*! \brief Task type */
   TaskType task_type;
   /*! \brief whether to average tree outputs */
-  bool average_tree_output;
+  bool average_tree_output{false};
   /*! \brief Group of parameters that are specific to the particular task type */
-  TaskParameter task_param;
+  TaskParam task_param{};
   /*! \brief extra parameters */
-  ModelParam param;
+  ModelParam param{};
+
+ protected:
+  /* Note: the following member fields shall be re-computed at serialization time */
+  // Number of trees
+  uint64_t num_tree_{0};
+  // Number of optional fields in the extension slot
+  int32_t num_opt_field_per_model_{0};
+  // Which Treelite version produced this model
+  int32_t major_ver_;
+  int32_t minor_ver_;
+  int32_t patch_ver_;
 
  private:
-  int major_ver_, minor_ver_, patch_ver_;
-  TypeInfo threshold_type_;
-  TypeInfo leaf_output_type_;
+  TypeInfo threshold_type_{TypeInfo::kInvalid};
+  TypeInfo leaf_output_type_{TypeInfo::kInvalid};
   // Internal functions for serialization
   virtual void GetPyBuffer(std::vector<PyBufferFrame>* dest) = 0;
   virtual void SerializeToFileImpl(FILE* dest_fp) = 0;
-  virtual void InitFromPyBuffer(std::vector<PyBufferFrame>::iterator begin,
-                                std::vector<PyBufferFrame>::iterator end) = 0;
+  // Load a Model object from a sequence of PyBuffer frames
+  // Returns the updated position of the cursor in the sequence
+  virtual std::vector<PyBufferFrame>::iterator InitFromPyBuffer(
+    std::vector<PyBufferFrame>::iterator it, std::size_t num_frame) = 0;
   virtual void DeserializeFromFileImpl(FILE* src_fp) = 0;
   template <typename HeaderPrimitiveFieldHandlerFunc>
   inline void SerializeTemplate(HeaderPrimitiveFieldHandlerFunc header_primitive_field_handler);
   template <typename HeaderPrimitiveFieldHandlerFunc>
   inline static void DeserializeTemplate(
       HeaderPrimitiveFieldHandlerFunc header_primitive_field_handler,
+      int32_t& major_ver, int32_t& minor_ver, int32_t& patch_ver,
       TypeInfo& threshold_type, TypeInfo& leaf_output_type);
 };
 
@@ -712,7 +845,7 @@ class ModelImpl : public Model {
   ModelImpl(ModelImpl&&) noexcept = default;
   ModelImpl& operator=(ModelImpl&&) noexcept = default;
 
-  void ReferenceSerialize(dmlc::Stream* fo) const override;
+  void DumpAsJSON(std::ostream& fo, bool pretty_print) const override;
   inline std::size_t GetNumTree() const override {
     return trees.size();
   }
@@ -722,8 +855,10 @@ class ModelImpl : public Model {
 
   inline void GetPyBuffer(std::vector<PyBufferFrame>* dest) override;
   inline void SerializeToFileImpl(FILE* dest_fp) override;
-  inline void InitFromPyBuffer(std::vector<PyBufferFrame>::iterator begin,
-                               std::vector<PyBufferFrame>::iterator end) override;
+  // Load a ModelImpl object from a sequence of PyBuffer frames
+  // Returns the updated position of the cursor in the sequence
+  inline std::vector<PyBufferFrame>::iterator InitFromPyBuffer(
+    std::vector<PyBufferFrame>::iterator it, std::size_t num_frame) override;
   inline void DeserializeFromFileImpl(FILE* src_fp) override;
 
  private:
@@ -733,11 +868,13 @@ class ModelImpl : public Model {
       HeaderPrimitiveFieldHandlerFunc header_primitive_field_handler,
       HeaderCompositeFieldHandlerFunc header_composite_field_handler,
       TreeHandlerFunc tree_handler);
-  template <typename HeaderFieldHandlerFunc, typename TreeHandlerFunc>
+  template <typename HeaderFieldHandlerFunc, typename TreeHandlerFunc,
+      typename SkipOptFieldHandlerFunc>
   inline void DeserializeTemplate(
       size_t num_tree,
       HeaderFieldHandlerFunc header_field_handler,
-      TreeHandlerFunc tree_handler);
+      TreeHandlerFunc tree_handler,
+      SkipOptFieldHandlerFunc skip_opt_field_handler);
 };
 
 }  // namespace treelite
