@@ -1,13 +1,13 @@
 /*!
- * Copyright (c) 2017-2020 by Contributors
+ * Copyright (c) 2017-2021 by Contributors
  * \file builder.cc
  * \brief model builder frontend
  * \author Hyunsu Cho
  */
 
-#include <dmlc/registry.h>
 #include <treelite/frontend.h>
 #include <treelite/tree.h>
+#include <treelite/logging.h>
 #include <memory>
 #include <queue>
 
@@ -65,8 +65,6 @@ struct TreeDraft {
 namespace treelite {
 namespace frontend {
 
-DMLC_REGISTRY_FILE_TAG(builder);
-
 struct TreeBuilderImpl {
   TreeDraft tree;
   inline TreeBuilderImpl(TypeInfo threshold_type, TypeInfo leaf_output_type)
@@ -86,11 +84,11 @@ struct ModelBuilderImpl {
     : trees(), num_feature(num_feature), num_class(num_class),
       average_tree_output(average_tree_output), threshold_type(threshold_type),
       leaf_output_type(leaf_output_type), cfg() {
-    CHECK_GT(num_feature, 0) << "ModelBuilder: num_feature must be positive";
-    CHECK_GT(num_class, 0) << "ModelBuilder: num_class must be positive";
-    CHECK(threshold_type != TypeInfo::kInvalid)
+    TREELITE_CHECK_GT(num_feature, 0) << "ModelBuilder: num_feature must be positive";
+    TREELITE_CHECK_GT(num_class, 0) << "ModelBuilder: num_class must be positive";
+    TREELITE_CHECK(threshold_type != TypeInfo::kInvalid)
       << "ModelBuilder: threshold_type can't be invalid";
-    CHECK(leaf_output_type != TypeInfo::kInvalid)
+    TREELITE_CHECK(leaf_output_type != TypeInfo::kInvalid)
       << "ModelBuilder: leaf_output_type can't be invalid";
   }
   // Templatized implementation of CommitModel()
@@ -106,7 +104,7 @@ void SetLeafVector(Tree<ThresholdType, LeafOutputType>* tree, int nid,
   std::vector<LeafOutputType> out_leaf_vector;
   for (size_t i = 0; i < leaf_vector_size; ++i) {
     const Value& leaf_value = leaf_vector[i];
-    CHECK(leaf_value.GetValueType() == expected_leaf_type)
+    TREELITE_CHECK(leaf_value.GetValueType() == expected_leaf_type)
       << "Leaf value at index " << i << " has incorrect type. Expected: "
       << TypeInfoToString(expected_leaf_type) << ", Given: "
       << TypeInfoToString(leaf_value.GetValueType());
@@ -132,7 +130,7 @@ class CreateHandle {
  public:
   inline static std::shared_ptr<void> Dispatch(const void* init_value) {
     const auto* v_ptr = static_cast<const ValueType*>(init_value);
-    CHECK(v_ptr);
+    TREELITE_CHECK(v_ptr);
     ValueType v = *v_ptr;
     return std::make_shared<ValueType>(v);
   }
@@ -141,7 +139,7 @@ class CreateHandle {
 Value
 Value::Create(const void* init_value, TypeInfo type) {
   Value value;
-  CHECK(type != TypeInfo::kInvalid) << "Type must be valid";
+  TREELITE_CHECK(type != TypeInfo::kInvalid) << "Type must be valid";
   value.type_ = type;
   value.handle_ = DispatchWithTypeInfo<CreateHandle>(type, init_value);
   return value;
@@ -150,18 +148,18 @@ Value::Create(const void* init_value, TypeInfo type) {
 template <typename T>
 T&
 Value::Get() {
-  CHECK(handle_);
+  TREELITE_CHECK(handle_);
   T* out = static_cast<T*>(handle_.get());
-  CHECK(out);
+  TREELITE_CHECK(out);
   return *out;
 }
 
 template <typename T>
 const T&
 Value::Get() const {
-  CHECK(handle_);
+  TREELITE_CHECK(handle_);
   const T* out = static_cast<const T*>(handle_.get());
-  CHECK(out);
+  TREELITE_CHECK(out);
   return *out;
 }
 
@@ -177,7 +175,8 @@ TreeBuilder::~TreeBuilder() = default;
 void
 TreeBuilder::CreateNode(int node_key) {
   auto& nodes = pimpl_->tree.nodes;
-  CHECK_EQ(nodes.count(node_key), 0) << "CreateNode: nodes with duplicate keys are not allowed";
+  TREELITE_CHECK_EQ(nodes.count(node_key), 0)
+    << "CreateNode: nodes with duplicate keys are not allowed";
   nodes[node_key] = std::make_unique<NodeDraft>();
 }
 
@@ -185,7 +184,7 @@ void
 TreeBuilder::DeleteNode(int node_key) {
   auto& tree = pimpl_->tree;
   auto& nodes = tree.nodes;
-  CHECK_GT(nodes.count(node_key), 0) << "DeleteNode: no node found with node_key";
+  TREELITE_CHECK_GT(nodes.count(node_key), 0) << "DeleteNode: no node found with node_key";
   NodeDraft* node = nodes[node_key].get();
   if (tree.root == node) {  // deleting root
     tree.root = nullptr;
@@ -208,9 +207,9 @@ void
 TreeBuilder::SetRootNode(int node_key) {
   auto& tree = pimpl_->tree;
   auto& nodes = tree.nodes;
-  CHECK_GT(nodes.count(node_key), 0) << "SetRootNode: no node found with node_key";
+  TREELITE_CHECK_GT(nodes.count(node_key), 0) << "SetRootNode: no node found with node_key";
   NodeDraft* node = nodes[node_key].get();
-  CHECK(!node->parent) << "SetRootNode: a root node cannot have a parent";
+  TREELITE_CHECK(!node->parent) << "SetRootNode: a root node cannot have a parent";
   tree.root = node;
 }
 
@@ -218,7 +217,7 @@ void
 TreeBuilder::SetNumericalTestNode(int node_key, unsigned feature_id, const char* opname,
                                   Value threshold, bool default_left, int left_child_key,
                                   int right_child_key) {
-  CHECK_GT(optable.count(opname), 0) << "No operator \"" << opname << "\" exists";
+  TREELITE_CHECK_GT(optable.count(opname), 0) << "No operator \"" << opname << "\" exists";
   Operator op = optable.at(opname);
   SetNumericalTestNode(node_key, feature_id, op, std::move(threshold), default_left,
                        left_child_key, right_child_key);
@@ -229,25 +228,26 @@ TreeBuilder::SetNumericalTestNode(int node_key, unsigned feature_id, Operator op
                                   bool default_left, int left_child_key, int right_child_key) {
   auto& tree = pimpl_->tree;
   auto& nodes = tree.nodes;
-  CHECK(tree.threshold_type == threshold.GetValueType())
+  TREELITE_CHECK(tree.threshold_type == threshold.GetValueType())
     << "SetNumericalTestNode: threshold has an incorrect type. "
     << "Expected: " << TypeInfoToString(tree.threshold_type)
     << ", Given: " << TypeInfoToString(threshold.GetValueType());
-  CHECK_GT(nodes.count(node_key), 0) << "SetNumericalTestNode: no node found with node_key";
-  CHECK_GT(nodes.count(left_child_key), 0)
+  TREELITE_CHECK_GT(nodes.count(node_key), 0)
+    << "SetNumericalTestNode: no node found with node_key";
+  TREELITE_CHECK_GT(nodes.count(left_child_key), 0)
     << "SetNumericalTestNode: no node found with left_child_key";
-  CHECK_GT(nodes.count(right_child_key), 0)
+  TREELITE_CHECK_GT(nodes.count(right_child_key), 0)
     << "SetNumericalTestNode: no node found with right_child_key";
   NodeDraft* node = nodes[node_key].get();
   NodeDraft* left_child = nodes[left_child_key].get();
   NodeDraft* right_child = nodes[right_child_key].get();
-  CHECK(node->status == NodeDraft::Status::kEmpty)
+  TREELITE_CHECK(node->status == NodeDraft::Status::kEmpty)
     << "SetNumericalTestNode: cannot modify a non-empty node";
-  CHECK(!left_child->parent)
+  TREELITE_CHECK(!left_child->parent)
     << "SetNumericalTestNode: node designated as left child already has a parent";
-  CHECK(!right_child->parent)
+  TREELITE_CHECK(!right_child->parent)
     << "SetNumericalTestNode: node designated as right child already has a parent";
-  CHECK(left_child != tree.root && right_child != tree.root)
+  TREELITE_CHECK(left_child != tree.root && right_child != tree.root)
     << "SetNumericalTestNode: the root node cannot be a child";
   node->status = NodeDraft::Status::kNumericalTest;
   node->left_child = nodes[left_child_key].get();
@@ -266,21 +266,22 @@ TreeBuilder::SetCategoricalTestNode(int node_key, unsigned feature_id,
                                     int left_child_key, int right_child_key) {
   auto &tree = pimpl_->tree;
   auto &nodes = tree.nodes;
-  CHECK_GT(nodes.count(node_key), 0) << "SetCategoricalTestNode: no node found with node_key";
-  CHECK_GT(nodes.count(left_child_key), 0)
+  TREELITE_CHECK_GT(nodes.count(node_key), 0)
+    << "SetCategoricalTestNode: no node found with node_key";
+  TREELITE_CHECK_GT(nodes.count(left_child_key), 0)
     << "SetCategoricalTestNode: no node found with left_child_key";
-  CHECK_GT(nodes.count(right_child_key), 0)
+  TREELITE_CHECK_GT(nodes.count(right_child_key), 0)
     << "SetCategoricalTestNode: no node found with right_child_key";
   NodeDraft* node = nodes[node_key].get();
   NodeDraft* left_child = nodes[left_child_key].get();
   NodeDraft* right_child = nodes[right_child_key].get();
-  CHECK(node->status == NodeDraft::Status::kEmpty)
+  TREELITE_CHECK(node->status == NodeDraft::Status::kEmpty)
     << "SetCategoricalTestNode: cannot modify a non-empty node";
-  CHECK(!left_child->parent)
+  TREELITE_CHECK(!left_child->parent)
     << "SetCategoricalTestNode: node designated as left child already has a parent";
-  CHECK(!right_child->parent)
+  TREELITE_CHECK(!right_child->parent)
     << "SetCategoricalTestNode: node designated as right child already has a parent";
-  CHECK(left_child != tree.root && right_child != tree.root)
+  TREELITE_CHECK(left_child != tree.root && right_child != tree.root)
     << "SetCategoricalTestNode: the root node cannot be a child";
   node->status = NodeDraft::Status::kCategoricalTest;
   node->left_child = nodes[left_child_key].get();
@@ -296,13 +297,14 @@ void
 TreeBuilder::SetLeafNode(int node_key, Value leaf_value) {
   auto& tree = pimpl_->tree;
   auto& nodes = tree.nodes;
-  CHECK(tree.leaf_output_type == leaf_value.GetValueType())
+  TREELITE_CHECK(tree.leaf_output_type == leaf_value.GetValueType())
     << "SetLeafNode: leaf_value has an incorrect type. "
     << "Expected: " << TypeInfoToString(tree.leaf_output_type)
     << ", Given: " << TypeInfoToString(leaf_value.GetValueType());
-  CHECK_GT(nodes.count(node_key), 0) << "SetLeafNode: no node found with node_key";
+  TREELITE_CHECK_GT(nodes.count(node_key), 0) << "SetLeafNode: no node found with node_key";
   NodeDraft* node = nodes[node_key].get();
-  CHECK(node->status == NodeDraft::Status::kEmpty) << "SetLeafNode: cannot modify a non-empty node";
+  TREELITE_CHECK(node->status == NodeDraft::Status::kEmpty)
+    << "SetLeafNode: cannot modify a non-empty node";
   node->status = NodeDraft::Status::kLeaf;
   node->leaf_value = std::move(leaf_value);
 }
@@ -314,14 +316,15 @@ TreeBuilder::SetLeafVectorNode(int node_key, const std::vector<Value>& leaf_vect
   const size_t leaf_vector_len = leaf_vector.size();
   for (size_t i = 0; i < leaf_vector_len; ++i) {
     const Value& leaf_value = leaf_vector[i];
-    CHECK(tree.leaf_output_type == leaf_value.GetValueType())
+    TREELITE_CHECK(tree.leaf_output_type == leaf_value.GetValueType())
       << "SetLeafVectorNode: the element " << i << " in leaf_vector has an incorrect type. "
       << "Expected: " << TypeInfoToString(tree.leaf_output_type)
       << ", Given: " << TypeInfoToString(leaf_value.GetValueType());
   }
-  CHECK_GT(nodes.count(node_key), 0) << "SetLeafVectorNode: no node found with node_key";
+  TREELITE_CHECK_GT(nodes.count(node_key), 0)
+    << "SetLeafVectorNode: no node found with node_key";
   NodeDraft* node = nodes[node_key].get();
-  CHECK(node->status == NodeDraft::Status::kEmpty)
+  TREELITE_CHECK(node->status == NodeDraft::Status::kEmpty)
     << "SetLeafVectorNode: cannot modify a non-empty node";
   node->status = NodeDraft::Status::kLeaf;
   node->leaf_vector = leaf_vector;
@@ -341,15 +344,15 @@ ModelBuilder::SetModelParam(const char* name, const char* value) {
 int
 ModelBuilder::InsertTree(TreeBuilder* tree_builder, int index) {
   if (tree_builder == nullptr) {
-    LOG(FATAL) << "InsertTree: not a valid tree builder";
+    TREELITE_LOG(FATAL) << "InsertTree: not a valid tree builder";
     return -1;
   }
   if (tree_builder->ensemble_id_ != nullptr) {
-    LOG(FATAL) << "InsertTree: tree is already part of another ensemble";
+    TREELITE_LOG(FATAL) << "InsertTree: tree is already part of another ensemble";
     return -1;
   }
   if (tree_builder->pimpl_->tree.threshold_type != this->pimpl_->threshold_type) {
-    LOG(FATAL)
+    TREELITE_LOG(FATAL)
       << "InsertTree: cannot insert the tree into the ensemble, because the ensemble requires all "
       << "member trees to use " << TypeInfoToString(this->pimpl_->threshold_type)
       << " type for split thresholds whereas the tree is using "
@@ -357,7 +360,7 @@ ModelBuilder::InsertTree(TreeBuilder* tree_builder, int index) {
     return -1;
   }
   if (tree_builder->pimpl_->tree.leaf_output_type != this->pimpl_->leaf_output_type) {
-    LOG(FATAL)
+    TREELITE_LOG(FATAL)
       << "InsertTree: cannot insert the tree into the ensemble, because the ensemble requires all "
       << "member trees to use " << TypeInfoToString(this->pimpl_->leaf_output_type)
       << " type for leaf outputs whereas the tree is using "
@@ -372,8 +375,9 @@ ModelBuilder::InsertTree(TreeBuilder* tree_builder, int index) {
         status == NodeDraft::Status::kCategoricalTest) {
       const int fid = static_cast<int>(kv.second->feature_id);
       if (fid < 0 || fid >= this->pimpl_->num_feature) {
-        LOG(FATAL) << "InsertTree: tree has an invalid split at node "
-                   << kv.first << ": feature id " << kv.second->feature_id << " is out of bound";
+        TREELITE_LOG(FATAL) << "InsertTree: tree has an invalid split at node "
+                            << kv.first << ": feature id "
+                            << kv.second->feature_id << " is out of bound";
         return -1;
       }
     }
@@ -391,7 +395,7 @@ ModelBuilder::InsertTree(TreeBuilder* tree_builder, int index) {
       tree_builder->ensemble_id_ = this;
       return index;
     } else {
-      LOG(FATAL) << "InsertTree: index out of bound";
+      TREELITE_LOG(FATAL) << "InsertTree: index out of bound";
       return -1;
     }
   }
@@ -410,7 +414,8 @@ ModelBuilder::GetTree(int index) const {
 void
 ModelBuilder::DeleteTree(int index) {
   auto& trees = pimpl_->trees;
-  CHECK_LT(static_cast<size_t>(index), trees.size()) << "DeleteTree: index out of bound";
+  TREELITE_CHECK_LT(static_cast<size_t>(index), trees.size())
+    << "DeleteTree: index out of bound";
   trees.erase(trees.begin() + index);
 }
 
@@ -430,7 +435,7 @@ ModelBuilderImpl::CommitModelImpl(ModelImpl<ThresholdType, LeafOutputType>* out_
   ModelImpl<ThresholdType, LeafOutputType>& model = *out_model;
   model.num_feature = this->num_feature;
   model.average_tree_output = this->average_tree_output;
-  model.task_param.output_type = TaskParameter::OutputType::kFloat;
+  model.task_param.output_type = TaskParam::OutputType::kFloat;
   model.task_param.num_class = this->num_class;
   // extra parameters
   InitParamAndCheck(&model.param, this->cfg);
@@ -443,8 +448,8 @@ ModelBuilderImpl::CommitModelImpl(ModelImpl<ThresholdType, LeafOutputType>* out_
 
   for (const auto& tree_builder : this->trees) {
     const auto& _tree = tree_builder.pimpl_->tree;
-    CHECK(_tree.root) << "CommitModel: a tree has no root node";
-    CHECK(_tree.root->status != NodeDraft::Status::kEmpty)
+    TREELITE_CHECK(_tree.root) << "CommitModel: a tree has no root node";
+    TREELITE_CHECK(_tree.root->status != NodeDraft::Status::kEmpty)
       << "SetRootNode: cannot set an empty node as root";
     model.trees.emplace_back();
     Tree<ThresholdType, LeafOutputType>& tree = model.trees.back();
@@ -459,15 +464,17 @@ ModelBuilderImpl::CommitModelImpl(ModelImpl<ThresholdType, LeafOutputType>* out_
       int nid;
       std::tie(node, nid) = Q.front();
       Q.pop();
-      CHECK(node->status != NodeDraft::Status::kEmpty)
+      TREELITE_CHECK(node->status != NodeDraft::Status::kEmpty)
         << "CommitModel: encountered an empty node in the middle of a tree";
       if (node->status == NodeDraft::Status::kNumericalTest) {
-        CHECK(node->left_child) << "CommitModel: a test node lacks a left child";
-        CHECK(node->right_child) << "CommitModel: a test node lacks a right child";
-        CHECK(node->left_child->parent == node) << "CommitModel: left child has wrong parent";
-        CHECK(node->right_child->parent == node) << "CommitModel: right child has wrong parent";
+        TREELITE_CHECK(node->left_child) << "CommitModel: a test node lacks a left child";
+        TREELITE_CHECK(node->right_child) << "CommitModel: a test node lacks a right child";
+        TREELITE_CHECK(node->left_child->parent == node)
+          << "CommitModel: left child has wrong parent";
+        TREELITE_CHECK(node->right_child->parent == node)
+          << "CommitModel: right child has wrong parent";
         tree.AddChilds(nid);
-        CHECK(node->threshold.GetValueType() == TypeToInfo<ThresholdType>())
+        TREELITE_CHECK(node->threshold.GetValueType() == TypeToInfo<ThresholdType>())
           << "CommitModel: The specified threshold has incorrect type. Expected: "
           << TypeInfoToString(TypeToInfo<ThresholdType>())
           << " Given: " << TypeInfoToString(node->threshold.GetValueType());
@@ -476,33 +483,35 @@ ModelBuilderImpl::CommitModelImpl(ModelImpl<ThresholdType, LeafOutputType>* out_
         Q.push({node->left_child, tree.LeftChild(nid)});
         Q.push({node->right_child, tree.RightChild(nid)});
       } else if (node->status == NodeDraft::Status::kCategoricalTest) {
-        CHECK(node->left_child) << "CommitModel: a test node lacks a left child";
-        CHECK(node->right_child) << "CommitModel: a test node lacks a right child";
-        CHECK(node->left_child->parent == node) << "CommitModel: left child has wrong parent";
-        CHECK(node->right_child->parent == node) << "CommitModel: right child has wrong parent";
+        TREELITE_CHECK(node->left_child) << "CommitModel: a test node lacks a left child";
+        TREELITE_CHECK(node->right_child) << "CommitModel: a test node lacks a right child";
+        TREELITE_CHECK(node->left_child->parent == node)
+          << "CommitModel: left child has wrong parent";
+        TREELITE_CHECK(node->right_child->parent == node)
+          << "CommitModel: right child has wrong parent";
         tree.AddChilds(nid);
         tree.SetCategoricalSplit(nid, node->feature_id, node->default_left, node->left_categories,
                                  false);
         Q.push({node->left_child, tree.LeftChild(nid)});
         Q.push({node->right_child, tree.RightChild(nid)});
       } else {  // leaf node
-        CHECK(node->left_child == nullptr && node->right_child == nullptr)
+        TREELITE_CHECK(node->left_child == nullptr && node->right_child == nullptr)
           << "CommitModel: a leaf node cannot have children";
         if (!node->leaf_vector.empty()) {  // leaf vector exists
-          CHECK_NE(flag_leaf_vector, 0)
+          TREELITE_CHECK_NE(flag_leaf_vector, 0)
             << "CommitModel: Inconsistent use of leaf vector: if one leaf node uses a leaf vector, "
             << "*every* leaf node must use a leaf vector";
           flag_leaf_vector = 1;  // now every leaf must use leaf vector
-          CHECK_EQ(node->leaf_vector.size(), model.task_param.num_class)
+          TREELITE_CHECK_EQ(node->leaf_vector.size(), model.task_param.num_class)
             << "CommitModel: The length of leaf vector must be identical to the number of output "
             << "groups";
           SetLeafVector(&tree, nid, node->leaf_vector);
         } else {  // ordinary leaf
-          CHECK_NE(flag_leaf_vector, 1)
+          TREELITE_CHECK_NE(flag_leaf_vector, 1)
             << "CommitModel: Inconsistent use of leaf vector: if one leaf node does not use a leaf "
             << "vector, *no other* leaf node can use a leaf vector";
           flag_leaf_vector = 0;  // now no leaf can use leaf vector
-          CHECK(node->leaf_value.GetValueType() == TypeToInfo<LeafOutputType>())
+          TREELITE_CHECK(node->leaf_value.GetValueType() == TypeToInfo<LeafOutputType>())
             << "CommitModel: The specified leaf value has incorrect type. Expected: "
             << TypeInfoToString(TypeToInfo<LeafOutputType>())
             << " Given: " << TypeInfoToString(node->leaf_value.GetValueType());
@@ -518,7 +527,7 @@ ModelBuilderImpl::CommitModelImpl(ModelImpl<ThresholdType, LeafOutputType>* out_
       // multi-class classifier, XGBoost/LightGBM style
       model.task_type = TaskType::kMultiClfGrovePerClass;
       model.task_param.grove_per_class = true;
-      CHECK_EQ(this->trees.size() % model.task_param.num_class, 0)
+      TREELITE_CHECK_EQ(this->trees.size() % model.task_param.num_class, 0)
         << "For multi-class classifiers with gradient boosted trees, the number of trees must be "
         << "evenly divisible by the number of output groups";
     } else {
@@ -530,10 +539,11 @@ ModelBuilderImpl::CommitModelImpl(ModelImpl<ThresholdType, LeafOutputType>* out_
     // multi-class classifier, sklearn RF style
     model.task_type = TaskType::kMultiClfProbDistLeaf;
     model.task_param.grove_per_class = false;
-    CHECK_GT(model.task_param.num_class, 1) << "Expected leaf vectors with length exceeding 1";
+    TREELITE_CHECK_GT(model.task_param.num_class, 1)
+      << "Expected leaf vectors with length exceeding 1";
     model.task_param.leaf_vector_size = model.task_param.num_class;
   } else {
-    LOG(FATAL) << "Impossible thing happened: model has no leaf node!";
+    TREELITE_LOG(FATAL) << "Impossible thing happened: model has no leaf node!";
   }
 }
 
