@@ -8,7 +8,6 @@
 
 #include "xgboost/xgboost_json.h"
 
-#include <fmt/format.h>
 #include <rapidjson/error/en.h>
 #include <rapidjson/document.h>
 #include <rapidjson/filereadstream.h>
@@ -313,8 +312,8 @@ bool RegTreeHandler::EndObject(std::size_t) {
 bool GBTreeModelHandler::StartArray() {
   return (push_key_handler<ArrayHandler<treelite::Tree<float, float>, RegTreeHandler>,
                            std::vector<treelite::Tree<float, float>>>(
-                               "trees", output.trees) ||
-          push_key_handler<IgnoreHandler>("tree_info"));
+                               "trees", output.model->trees) ||
+          push_key_handler<ArrayHandler<int>, std::vector<int>>("tree_info", output.tree_info));
 }
 
 bool GBTreeModelHandler::StartObject() {
@@ -339,10 +338,9 @@ bool GradientBoosterHandler::String(const char *str,
   }
 }
 bool GradientBoosterHandler::StartObject() {
-  if (push_key_handler<GBTreeModelHandler, treelite::ModelImpl<float, float>>("model", output)) {
+  if (push_key_handler<GBTreeModelHandler, ParsedXGBoostModel>("model", output)) {
     return true;
-  } else if (push_key_handler<GradientBoosterHandler, treelite::ModelImpl<float, float>>("gbtree",
-                                                                                         output)) {
+  } else if (push_key_handler<GradientBoosterHandler, ParsedXGBoostModel>("gbtree", output)) {
     // "dart" booster contains a standard gbtree under ["gradient_booster"]["gbtree"]["model"].
     return true;
   } else {
@@ -357,11 +355,12 @@ bool GradientBoosterHandler::StartArray() {
 bool GradientBoosterHandler::EndObject(std::size_t memberCount) {
   if (name == "dart" && !weight_drop.empty()) {
     // Fold weight drop into leaf value for dart models.
-    TREELITE_CHECK_EQ(output.trees.size(), weight_drop.size());
-    for (size_t i = 0; i < output.trees.size(); ++i) {
-      for (int nid = 0; nid < output.trees[i].num_nodes; ++nid) {
-        if (output.trees[i].IsLeaf(nid)) {
-          output.trees[i].SetLeaf(nid, weight_drop[i] * output.trees[i].LeafValue(nid));
+    auto& trees = output.model->trees;
+    TREELITE_CHECK_EQ(trees.size(), weight_drop.size());
+    for (size_t i = 0; i < trees.size(); ++i) {
+      for (int nid = 0; nid < trees[i].num_nodes; ++nid) {
+        if (trees[i].IsLeaf(nid)) {
+          trees[i].SetLeaf(nid, weight_drop[i] * trees[i].LeafValue(nid));
         }
       }
     }
@@ -378,7 +377,9 @@ bool ObjectiveHandler::StartObject() {
           push_key_handler<IgnoreHandler>("tweedie_regression_param") ||
           push_key_handler<IgnoreHandler>("softmax_multiclass_param") ||
           push_key_handler<IgnoreHandler>("lambda_rank_param") ||
-          push_key_handler<IgnoreHandler>("aft_loss_param"));
+          push_key_handler<IgnoreHandler>("aft_loss_param") ||
+          push_key_handler<IgnoreHandler>("pseduo_huber_param") ||
+          push_key_handler<IgnoreHandler>("pseudo_huber_param"));
 }
 
 bool ObjectiveHandler::String(const char *str, std::size_t length, bool) {
@@ -404,7 +405,8 @@ bool LearnerParamHandler::String(const char *str,
                        output.param.global_bias) ||
           assign_value("num_class", static_cast<unsigned int>(std::max(std::atoi(str), 1)),
                        output.task_param.num_class) ||
-          assign_value("num_feature", std::atoi(str), output.num_feature));
+          assign_value("num_feature", std::atoi(str), output.num_feature) ||
+          check_cur_key("boost_from_average"));
 }
 
 /******************************************************************************
@@ -414,8 +416,8 @@ bool LearnerHandler::StartObject() {
   // "attributes" key is not documented in schema
   return (push_key_handler<LearnerParamHandler, treelite::ModelImpl<float, float>>(
               "learner_model_param", *output.model) ||
-          push_key_handler<GradientBoosterHandler, treelite::ModelImpl<float, float>>(
-              "gradient_booster", *output.model) ||
+          push_key_handler<GradientBoosterHandler, ParsedXGBoostModel>(
+              "gradient_booster", output) ||
           push_key_handler<ObjectiveHandler, std::string>("objective", objective) ||
           push_key_handler<IgnoreHandler>("attributes"));
 }
@@ -441,7 +443,7 @@ bool XGBoostCheckpointHandler::StartArray() {
 }
 
 bool XGBoostCheckpointHandler::StartObject() {
-  return push_key_handler<LearnerHandler, XGBoostModelHandle>("learner", output);
+  return push_key_handler<LearnerHandler, ParsedXGBoostModel>("learner", output);
 }
 
 /******************************************************************************
@@ -453,9 +455,9 @@ bool XGBoostModelHandler::StartArray() {
 }
 
 bool XGBoostModelHandler::StartObject() {
-  return (push_key_handler<LearnerHandler, XGBoostModelHandle>("learner", output) ||
+  return (push_key_handler<LearnerHandler, ParsedXGBoostModel>("learner", output) ||
           push_key_handler<IgnoreHandler>("Config") ||
-          push_key_handler<XGBoostCheckpointHandler, XGBoostModelHandle>("Model", output));
+          push_key_handler<XGBoostCheckpointHandler, ParsedXGBoostModel>("Model", output));
 }
 
 bool XGBoostModelHandler::EndObject(std::size_t memberCount) {
@@ -488,14 +490,13 @@ bool XGBoostModelHandler::EndObject(std::size_t memberCount) {
  * RootHandler
  * ***************************************************************************/
 bool RootHandler::StartObject() {
-  handle = {dynamic_cast<treelite::ModelImpl<float, float>*>(output.get()), {}, ""};
-  return push_handler<XGBoostModelHandler, XGBoostModelHandle>(handle);
+  return push_handler<XGBoostModelHandler, ParsedXGBoostModel>(output);
 }
 
 /******************************************************************************
  * DelegatedHandler
  * ***************************************************************************/
-std::unique_ptr<treelite::Model> DelegatedHandler::get_result() { return std::move(result); }
+ParsedXGBoostModel DelegatedHandler::get_result() { return std::move(result); }
 bool DelegatedHandler::Null() { return delegates.top()->Null(); }
 bool DelegatedHandler::Bool(bool b) { return delegates.top()->Bool(b); }
 bool DelegatedHandler::Int(int i) { return delegates.top()->Int(i); }
@@ -540,6 +541,37 @@ std::unique_ptr<treelite::Model> ParseStream(std::unique_ptr<StreamType> input_s
                         << "Parsing error at offset " << offset << ": "
                         << rapidjson::GetParseError_En(error_code) << "\n" << diagnostic;
   }
-  return handler->get_result();
+  treelite::details::ParsedXGBoostModel parsed = handler->get_result();
+
+  // Special handling for multi-class classifier with num_parallel_tree > 1
+  if (parsed.model->task_param.grove_per_class && parsed.model->task_param.num_class > 2) {
+    // Infer num_parallel_tree
+    unsigned num_parallel_tree = 0;
+    for (int e : parsed.tree_info) {
+      if (e != 0) {
+        break;
+      }
+      ++num_parallel_tree;
+    }
+    if (num_parallel_tree > 1) {
+      // Re-order trees to recover the grove-per-class layout.
+      // The prediction for the i-th class is determined by the trees whose index is congruent
+      // to [i] modulo [num_class].
+      // Currently, the trees' association with classes is as follows:
+      // 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2
+      // We need to re-order them as follows:
+      // 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2
+      std::vector<treelite::Tree<float, float>> new_trees;
+      std::size_t num_tree = parsed.model->trees.size();
+      for (std::size_t c = 0; c < num_parallel_tree; ++c) {
+        for (std::size_t tree_id = c; tree_id < num_tree; tree_id += num_parallel_tree) {
+          new_trees.push_back(std::move(parsed.model->trees[tree_id]));
+        }
+      }
+      TREELITE_CHECK_EQ(new_trees.size(), num_tree);
+      parsed.model->trees = std::move(new_trees);
+    }
+  }
+  return std::move(parsed.model_ptr);
 }
 }  // anonymous namespace
